import json

import pytest
import pkg_resources

from ..tasks import gracedb, superevent_manager


def test_set_preferred_event(monkeypatch):
    class T0212TTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/T0212_S0039_preferred.json') as f:
                return json.load(f)

    class NewEventTTPResponse(object):
        def json(self):
            return dict(graceid="T1234",
                        instruments="I1,J1,K1,L1,M1",
                        superevent="some_superevent",
                        far=1e-30)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self.__update_superevent_called = False

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                return NewEventTTPResponse()

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self.__update_superevent_called
            # function should be called only once
            self.__update_superevent_called = True

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    gracedb.set_preferred_event('S0039', 'T0212', 'T1234')


def test_get_superevent(monkeypatch):
    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def get(self, url):
            return SuperEventsTTPResponse()

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())

    gid_1 = 'T0212'  # present in fake superevent queue
    gid_2 = 'T0219'  # present but not preferred event
    gid_3 = 'T9999'  # absent in fake superevent queue
    superevent_id_1, preferred_flag_1, r_1 = gracedb.get_superevent(gid_1)
    superevent_id_2, preferred_flag_2, r_2 = gracedb.get_superevent(gid_2)
    superevent_id_3, preferred_flag_3, r_3 = gracedb.get_superevent(gid_3)

    assert superevent_id_1 == 'S0039' and preferred_flag_1
    assert superevent_id_2 == 'S0041' and not preferred_flag_2
    assert superevent_id_3 is None


def test_parse_trigger_raising():
    garbage_payload = dict(some_value=999, something_else=99)
    garbage_text = json.dumps(garbage_payload)
    with pytest.raises(KeyError):
        superevent_manager.superevent_handler(garbage_text)


# FIXME lot of redundancy, implement setUp tearDown for below
def test_parse_trigger_1(monkeypatch):
    """New trigger G000000, less significant than already
    existing superevent"""
    class T0212TTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/T0212_S0039_preferred.json') as f:
                return json.load(f)

    class NewEventTTPResponse(object):
        def json(self):
            # hardcoding values of the new trigger G000000
            return dict(graceid="G000000",
                        instruments="H1,L1",
                        superevent="S0039",
                        far=3.021362523404484e-09)

    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self._addevent_called = False

        def get(self, url):
            return SuperEventsTTPResponse()

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._addevent_called
            self._addevent_called = True

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                return NewEventTTPResponse()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    with pkg_resources.resource_stream(
            __name__, 'data/mock_trigger_new_G000000.json') as f:
        text = f.read()
    # New trigger G000000 time falls in S0039 window
    superevent_manager.superevent_handler(text)


def test_parse_trigger_2(monkeypatch):
    """New trigger G000000, more significant than already
    existing superevent"""
    class T0212TTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/T0212_S0039_preferred.json') as f:
                return json.load(f)

    class NewEventTTPResponse(object):
        def json(self):
            # hacking the far of G000000 to a very low value
            return dict(graceid="G000000",
                        instruments="H1,L1",
                        superevent="S0039",
                        far=3.021362523404484e-31)

    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self._update_superevent_called = False
            self._addevent_called = False

        def get(self, url):
            return SuperEventsTTPResponse()

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._addevent_called
            self._addevent_called = True

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._update_superevent_called
            # function should be called only once
            self._update_superevent_called = True

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                return NewEventTTPResponse()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    # New trigger G000000 falls in S0039 window
    with pkg_resources.resource_stream(
            __name__, 'data/mock_trigger_new_G000000.json') as f:
        text = f.read()
    superevent_manager.superevent_handler(text)


def test_parse_trigger_3(monkeypatch):
    """New trigger G000001, not present among superevents
    New superevent created"""
    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self._create_superevent_called = False

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._create_superevent_called
            self._create_superevent_called = True

        def get(self, url):
            return SuperEventsTTPResponse()

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    with pkg_resources.resource_stream(
            __name__, 'data/mock_trigger_new_G000001.json') as f:
        text = f.read()
    # G000001 absent in any superevent window, new superevent created
    superevent_manager.superevent_handler(text)


def test_parse_trigger_4(monkeypatch):
    """Restart scenario - G000001 not present, update alert
    received.
    Should give warning in captured log while running
    unittests"""
    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class G000001TTPResponse(object):
        def json(self):
            # FIXME json packet currently has only necessary keys
            return dict(graceid="G000001",
                        instruments="H1,L1",
                        gpstime=1286741861.52678,
                        far=3.021362523404484e-31)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self._create_superevent_called = False

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._create_superevent_called
            self._create_superevent_called = True

        def get(self, url):
            return SuperEventsTTPResponse()

        def event(self, gid):
            return G000001TTPResponse()

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    with pkg_resources.resource_stream(
            __name__, 'data/mock_trigger_update_G000001.json') as f:
        text = f.read()
    # G000001 absent in any superevent window, update alert
    # triggers creation of a new superevent
    superevent_manager.superevent_handler(text)


def test_parse_trigger_5(monkeypatch):
    """Update alert comes in for already existing superevent T0211"""
    class SuperEventsTTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/superevents.json') as f:
                return json.load(f)

    class T0212TTPResponse(object):
        def json(self):
            with pkg_resources.resource_stream(
                    __name__, 'data/T0212_S0039_preferred.json') as f:
                return json.load(f)

    class T0211TTPResponse(object):
        def json(self):
            # The update will change the preferred event
            return dict(graceid="T0211",
                        instruments="H1,L1",
                        gpstime=1163905217.52678,
                        far=5.500858049414366e-20)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'
            self._update_superevent_called = False

        def get(self, url):
            return SuperEventsTTPResponse()

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                return T0211TTPResponse()

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            assert not self._update_superevent_called
            self._update_superevent_called = True

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())
    with pkg_resources.resource_stream(
            __name__, 'data/mock_trigger_update_T0211.json') as f:
        text = f.read()
    superevent_manager.superevent_handler(text)



"""Module containing the functionality for creation and management of
superevents.

    * There is serial processing of triggers from low latency
      pipelines.
    * Dedicated **superevent** queue for this purpose.
    * Primary logic to respond to low latency triggers contained
      in :meth:`superevent_handler` function.
"""
import json

from celery.utils.log import get_task_logger

from ..celery import app
from . import gracedb, lvalert

log = get_task_logger(__name__)


@lvalert.handler('cbc_gstlal',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 'cbc_gstlal_mdc',
                 'test_gstlal',
                 'test_pycbc',
                 'test_mbtaonline',
                 'test_gstlal_mdc',
                 queue='superevent',
                 shared=False)
def superevent_handler(text):
    """LVAlert handler for superevent manager.
    Recieves payload from test and production nodes and
    serially processes them to create/modify superevents
    """
    payload = json.loads(text)

    gid = payload['uid']
    alert_type = payload['alert_type']
    gpstime = payload['object'].get('gpstime')
    # s_event_dict['superevents'] is the list of superevents,
    # sid is True if gid exists in any superevent gw_list
    sid, preferred_flag, s_event_dict = gracedb.get_superevent(gid)
    # FIXME as config grows, this could go into separate function
    d_t_start = app.conf['superevent_d_t_start']
    d_t_end = app.conf['superevent_d_t_end']
    # Condition 1/4
    if sid is None and alert_type == 'new':
        log.debug('Entered 1st if')
        # Check which superevent window trigger gpstime falls in
        for superevent in s_event_dict['superevents']:
            if superevent['t_start'] <= gpstime < superevent['t_end']:
                log.info('Event %s in window of %s',
                         gid, superevent['superevent_id'])
                gracedb.add_event_to_superevent(superevent['superevent_id'],
                                                gid)
                # ADDITIONAL LOGIC HERE IF THE TIME
                # WINDOW IS TO BE CHANGED BASED ON
                # TRIGGER PARAMETERS
                # gracedb is a module under tasks and not gracedb-client
                gracedb.set_preferred_event(superevent['superevent_id'],
                                            superevent['preferred_event'],
                                            gid)
                break
        # Create a new event if not in any time window
        else:
            gracedb.create_superevent(payload,
                                      d_t_start=d_t_start,
                                      d_t_end=d_t_end)

    # Condition 2/4
    elif sid is None and alert_type == 'update':
        # Probable case of a restart
        log.debug('2nd if: Possible restart scenario')
        log.warning('No superevent found for alert_type update for %s', gid)
        log.warning('Creating new superevent for %s', gid)
        # update alerts don't have gpstime, hence fetch from gracedb
        event_dict = gracedb.get_event(gid)
        # add gpstime to payload dict since update alert don't have gpstime
        payload['object']['gpstime'] = event_dict['gpstime']

        gracedb.create_superevent(payload,
                                  d_t_start=d_t_start,
                                  d_t_end=d_t_end)

    # Condition 3/4
    elif sid and alert_type == 'new':
        # ERROR SITUATION
        log.debug('3rd if: SID returned and new alert')
        log.critical('Superevent %s exists for alert_type new for %s',
                     sid, gid)

    # Condition 4/4
    elif sid and alert_type == 'update':
        # Logic to update the preferred G event
        log.debug('4th if: Update the pointer')
        for entry in s_event_dict['superevents']:
            if entry['superevent_id'] == sid:
                break
        # if the preferred_event entry does not exists for some reason
        if not entry['preferred_event']:
            # FIXME: To be implemented
            raise NotImplementedError
        preferred_event = entry['preferred_event']
        # logic to decide if new event is preferred
        gracedb.set_preferred_event(entry['superevent_id'],
                                    preferred_event, gid)
    # Condition else
    else:
        log.critical('Unhandled by parse_trigger, passing...')



"""Routing of LVAlert messages to other tasks."""
import json
import os

from celery import group

from . import bayestar
from . import circulars
from . import gracedb
from . import lvalert
from . import skymaps
# from .voevent import send


@lvalert.handler('cbc_gstlal',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 'cbc_gstlal_mdc',
                 shared=False)
def dispatch(payload):
    """Parse an LVAlert message and dispatch it to other tasks."""
    # Parse JSON payload
    alert = json.loads(payload)

    # Determine GraceDB ID
    graceid = alert['uid']

    if alert['alert_type'] == 'update' and 'voevent_type' in alert['object']:
        # FIXME: temporarily disable sending GCNs as per P. Brady request
        pass  # send.delay(alert['object']['text'])
    elif alert['alert_type'] == 'update' and alert.get('file'):
        _, versioned_filename = os.path.split(alert['object']['file'])
        filename, _, _ = versioned_filename.rpartition(',')
        filebase, fitsext, _ = filename.rpartition('.fits')
        tags = alert['object']['tag_names']
        if fitsext:
            skymaps.annotate_fits(
                versioned_filename, filebase, graceid, tags).delay()
        elif filename == 'psd.xml.gz':
            group(
                bayestar.bayestar(graceid),
                circulars.create_circular.s(graceid) |
                gracedb.upload.s('circular.txt', graceid,
                                 'Automated circular')
            ).delay()



"""Communication with GraceDB."""
import os
import sys

from ligo.gracedb import rest
from celery.local import PromiseProxy
from celery.utils.log import get_task_logger

from ..celery import app

# Defer initializing the GraceDb REST client until it is needed,
# because if the user lacks valid credentials, then the API will
# raise an exception as soon as it is instantiated---which would
# otherwise make it impossible to import this module without first
# logging in.
#
# FIXME: sphinx gets confused by PromiseProxy and prints:
#     Recursion error:
#     maximum recursion depth exceeded
# so don't create the proxy if we are being run by sphinx.
prog = os.path.basename(sys.argv[0])
if prog != 'sphinx-build' and 'build_sphinx' not in sys.argv:
    client = PromiseProxy(rest.GraceDb,
                          ('https://' + app.conf.gracedb_host + '/api/',))

log = get_task_logger(__name__)


@app.task(shared=False)
def create_event(filecontents, search, pipeline, group):
    """Create an event in GraceDb."""
    response = client.createEvent(group=group, pipeline=pipeline,
                                  filename='initial.data', search=search,
                                  filecontents=filecontents)
    return response.json()['graceid']


@app.task(ignore_result=True, shared=False)
def create_tag(tag, n, graceid):
    """Create a tag in GraceDb."""
    client.createTag(graceid, n, tag)


@app.task(shared=False)
def download(filename, graceid):
    """Download a file from GraceDB."""
    return client.files(graceid, filename, raw=True).read()


@app.task(shared=False)
def get_events(query=None, orderby=None, count=None, columns=None):
    """Get events from GraceDb."""
    return list(client.events(query=query, orderby=orderby,
                count=count, columns=columns))


@app.task(shared=False)
def get_event(graceid):
    """Retrieve an event from GraceDb."""
    return client.event(graceid).json()


@app.task(shared=False)
def get_log(graceid):
    """Get all log messages for an event in GraceDb."""
    return client.logs(graceid).json()['log']


@app.task(shared=False)
def replace_event(graceid, payload):
    """Get an event from GraceDb."""
    client.replaceEvent(graceid, 'initial.data', filecontents=payload)


@app.task(ignore_result=True, shared=False)
def upload(filecontents, filename, graceid, message, tags=()):
    """Upload a file to GraceDB."""
    client.writeLog(graceid, message, filename, filecontents, tags)


@app.task(shared=False)
def get_superevent(gid):
    """Iterate through superevents in gracedb and return sid if
    gid exists in the association.

    Parameters
    ----------
    gid : str
        uid of the trigger to be checked

    Returns
    -------
    superevent_id : str
        uid of the superevent. None if not found
    preferred_flag : bool
        True if gid is found and it is preferred. None if not found.
    r : dict
        The JSON packet of the superevent. None if not found.
    """
    r = client.get(client.service_url+"superevents/")
    r_json = r.json()
    superevents = r_json['superevents']
    for superevent in superevents:
        preferred_flag = False
        # check preferred_event first
        if gid == superevent['preferred_event']:
            preferred_flag = True
            log.info('Found association (Preferred) %s <-> %s',
                     gid, superevent['superevent_id'])
            return superevent['superevent_id'], preferred_flag, r_json
        # then check the gw_events
        elif gid in superevent['gw_events']:
            log.info('Found association (NOT Preferred) %s <-> %s',
                     gid, superevent['superevent_id'])
            return superevent['superevent_id'], preferred_flag, r_json
    return None, False, r_json


@app.task(ignore_result=True, shared=False)
def set_preferred_event(sid, preferred_event, gid):
    """
    Update superevent with the new trigger id based on
    FAR values.

    Parameters
    ----------
    sid : str
        superevent uid
    preferred_event : str
        preferred event id of the superevent
    gid : str
        uid of the new trigger
    """
    r_new_event = client.event(gid).json()
    r_preferred_event = client.event(preferred_event).json()
    if r_new_event['far'] < r_preferred_event['far']:
        client.updateSuperevent(sid,
                                preferred_event=r_new_event['graceid'])


@app.task(ignore_result=True, shared=False)
def create_superevent(payload, d_t_start=5, d_t_end=5):
    """
    Create new superevent in GraceDb with preferred G event.
    """
    t0 = payload['object']['gpstime']
    gid = payload['uid']
    ts = t0 - d_t_start
    te = t0 + d_t_end
    client.createSuperevent(ts, t0, te, preferred_event=gid)
    log.info('Successfully created superevent')


@app.task(ignore_result=True, shared=False)
def add_event_to_superevent(sid, gid):
    """Wrapper for GraceDb.addEventToSuperevent.
    Can be called in async if required
    """
    client.addEventToSuperevent(sid, gid)
